(* 
 * Copyright (c) 2001 The University of Wroclaw.
 * All rights reserved. See file COPYRIGHT for details about licensing.
 *
 * $Id: gontcodegen.ml,v 1.23 2001/11/27 10:43:09 malekith Exp $
 *
 * Module implementation: Gontcodegen
 * Description: This module generally converts Gont.* -> Ksi.*. 
 *)

open Gontutil
open Gontbind
open Gonttypes

type fnc_ctx = {ctx_level: int;
                ctx_closure_type: id;
                ctx_aux_decls: Ksi.decl list;
                ctx_fnname: id;
                ctx_module: id}
                

type ctx = {gctx_globals: (ty * id) list;
            gctx_global_types: typedef list;
            gctx_module: id;
            gctx_file: string}

let global_ctx = {ctx_level = 0; 
                  ctx_closure_type = "";
                  ctx_aux_decls = [];
                  ctx_fnname = "???";
                  ctx_module = "???"}
;;

(* This is here mainly to remaind you, that you're changing namespace.
 * You go from Gont to Ksi names, which in most cases are the same,
 * but they surly doesn't have to. That's why Ksi.id type is hidden. 
 *)
let ksi_name s = 
  let e = String.copy s in
  for i = 0 to String.length e - 1 do
    if e.[i] = ':' then e.[i] <- '$'
  done;
  Ksi.make_id e

let lambda_exp = Ksi.Struct(ksi_name "lambda_exp$")
let ksi_closure = Ksi.Ref(ksi_name "cl$")
let ksi_closure_var n = Ksi.Arrow(ksi_closure, ksi_name n)
let ksi_parent_closure = Ksi.Ref(ksi_name "parent_cl$")
;;

let created_structs : ((ty * id) list, id) Hashtbl.t = Hashtbl.create 16
let created_unions : ((ty * id) list, id) Hashtbl.t = Hashtbl.create 16 
;;

let init_codegen m f =
  Hashtbl.clear created_structs;
  Hashtbl.clear created_unions;
  let (s, t) = Gontmodule.globals m in
  {gctx_module = m;
   gctx_globals = s;
   gctx_global_types = t;
   gctx_file = f}
;;

let global_check loc ctx sym =
  let rec check l =
    match l with
        (t, n)::xs when (ctx.gctx_module ^ "::" ^ n) = sym#get_name ->
           (try
              global_subtype t sym#get_type
            with Type_error e -> 
               err (loc, xf "type of %s is incompatibile with type \
                             declared in interface: %s" 
                             sym#get_name e));
           xs
      | x::xs ->
             x :: (check xs)
      | [] -> [] in
   let new_g = check ctx.gctx_globals in
     if List.length new_g = List.length ctx.gctx_globals then
       (ctx, [Ksi.Local])
     else
       ({ctx with gctx_globals = new_g}, [Ksi.Public])

let global_type_check loc ctx td =
  let rec check l =
    match l with
        ({td_name = n} as p)::xs when n = td.td_name ->
           (try
              match p.td_type with
                  AbstractType -> ()
                | _ -> global_subtype p.td_type td.td_type
            with Type_error e -> 
               err (loc, xf "type of %s is incompatibile with type \
                             declared in interface: %s" 
                             td.td_name e));
           xs
      | x::xs ->
             x :: (check xs)
      | [] -> [] in
   let new_g = check ctx.gctx_global_types in
     if List.length new_g = List.length ctx.gctx_globals then
       (ctx, [Ksi.Local])
     else
       ({ctx with gctx_global_types = new_g}, [Ksi.Public])
;;

let ksi_runtime_proto {gctx_module = m; gctx_file=f} =
  let fnc (r, n, a) = Ksi.FuncProto([Ksi.Extern], r, n, a) in
  let ptr = Ksi.Ptr(Ksi.Void) in
  let size_t = Ksi.uptroff () in
[
  Ksi.TopComment("implementation of module " ^ m);
  Ksi.InputFile(f);
  Ksi.TopComment("BEG Runtime proto");
  Ksi.StructDef(ksi_name "lambda_exp$", [], 
                [(ptr, ksi_name "fnc"); 
                 (ptr, ksi_name "closure")]);
  fnc(ptr, ksi_name "make_closure$", [size_t]);
  fnc(Ksi.Ptr(lambda_exp), ksi_name "make_lambda$", [ptr; ptr]);
  Ksi.StructDef(ksi_name "string$", [], 
                [(size_t, ksi_name "len"); 
                 (Ksi.Ptr(Ksi.UInt(8)), ksi_name "data")]);
  fnc(Ksi.Ptr(Ksi.Struct(ksi_name "string$")), 
      ksi_name "make_string$",
      [size_t; Ksi.Ptr(Ksi.UInt(8))]);
  fnc(Ksi.Ptr(Ksi.c_type Ksi.C_double), 
      ksi_name "make_float$",
      [Ksi.c_type Ksi.C_double]);
  Ksi.TopComment("END Runtime proto")
]
;;

let aux_decls : Ksi.decl list ref = ref []

let rec ksi_struct l =
  try
    Hashtbl.find created_structs l
  with Not_found ->
    let seq = (seq_name "struct") ^ "$" in
    let struct_decl =
        Ksi.StructDef(ksi_name seq, [], 
                      List.map (fun (t, n) -> (ksi_local_type t, 
                                               ksi_name n))
                               l) in
    Hashtbl.add created_structs l seq;
    aux_decls := struct_decl :: !aux_decls; 
    seq
and ksi_union l =
  try
    Hashtbl.find created_unions l
  with Not_found ->
    let seq = (seq_name "union") ^ "$" in
    let union_decl =
        Ksi.UnionDef(ksi_name (seq ^ "content"), [], 
                     List.map (fun (t, n) -> (ksi_local_type t, 
                                              ksi_name n))
                              l) in
    let struct_decl =
        Ksi.StructDef(ksi_name seq, [],
                      [(Ksi.uptroff(), ksi_name "sel");
                       (Ksi.Struct(ksi_name (seq ^ "content")), 
                        ksi_name "u")]) in
    Hashtbl.add created_unions l seq;
    aux_decls := struct_decl :: !aux_decls; 
    seq

(* Return Ksi type, to be used as type of local variable.
 * This is sometimes not exactly, what is passed to and returned from
 * functions, when called. *)
and ksi_local_type t =
  match t with
      Int | Bool -> Ksi.ptroff ()
    | AbstractType | Void -> 
       raise (Error "cannot have variable of abstract or void type")
    | Float -> Ksi.c_type(Ksi.C_double)
    | Func _ -> Ksi.Ptr(Ksi.Struct(ksi_name "lambda_exp$"))
    | String -> Ksi.Ptr(Ksi.Struct(ksi_name "string$"))
    | TyVar(_) -> raise (Error "cannot have variable of unspecified type")
    | Tuple(_) -> Ksi.Ptr(Ksi.Ptr(Ksi.Void))
    | Struct(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | Union(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_union l)))
    | OptStruct(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | NamedType _ -> ksi_local_type (instance_value t)

(* Return ksi type, to be used when passing arguments, returning values. *)
and ksi_passed_type t =
  match t with
      Int | Bool -> Ksi.ptroff ()
    | Void -> Ksi.Void
    | Float -> Ksi.Ptr(Ksi.c_type(Ksi.C_double))
    | Func _ -> Ksi.Ptr(Ksi.Struct(ksi_name "lambda_exp$"))
    | String -> Ksi.Ptr(Ksi.Struct(ksi_name "string$"))
    | TyVar(_) -> Ksi.Ptr(Ksi.Void)
    | Tuple(_) -> Ksi.Ptr(Ksi.Ptr(Ksi.Void))
    | Struct(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | Union(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_union l)))
    | OptStruct(l) -> Ksi.Ptr(Ksi.Struct(ksi_name (ksi_struct l)))
    | NamedType _ -> ksi_passed_type (instance_value t)
    | AbstractType -> Ksi.Ptr(Ksi.Void)
;;

let ksi_op2 o =
  match o with
      IntPlus | FloatPlus -> Ksi.Plus
    | IntMinus | FloatMinus -> Ksi.Minus
    | IntMult | FloatMult -> Ksi.Mult
    | IntDiv -> Ksi.TruncDiv
    | IntMod -> Ksi.TruncMod
    | FloatDiv -> Ksi.RDiv
    | Lt -> Ksi.Lt
    | Gt -> Ksi.Gt
    | Le -> Ksi.Le
    | Ge -> Ksi.Ge
    | Eq -> Ksi.Eq
    | Ne -> Ksi.Ne
    | Assign -> assert false
  
let ksi_op1 o =
  match o with
      IntNegate | FloatNegate -> Ksi.Negate
;;

let rec ksi_stmt (c, s) =
  let (c', s') =
    match s.raw_stmt with
        Skip ->
          (c, Ksi.Nop)
      | Exp(e) ->
          ksi_exp (c, e)
      | Return(e) ->
          let (c', e') = ksi_exp (c, e) in
          let e'' = 
            match type_of e with
                Float -> Ksi.Call(ksi_name "make_float$", [e'])
              | _ -> e' in
          (c', Ksi.Return(e''))
      | ReturnVoid ->
          (c, Ksi.Return(Ksi.Nop))
      | Block([]) ->
          (c, Ksi.Nop)
      | Block([s]) ->
          let (c', s') = ksi_stmt (c, s) in (c', Ksi.Bind(s'))
      | Block([s1; s2]) ->
          let (c1, s1') = ksi_stmt (c, s1) in
          let (c2, s2') = ksi_stmt (c1, s2) in
          (c2, Ksi.Bind(Ksi.Compound(s1', s2')))
      | Block(s) ->
          let (c', s') = list_apply ksi_stmt (c, s) in
          (c', Ksi.Bind(Ksi.CompoundList(s')))
      | VarDef v ->
          (match v.var_symbol with
              Some sym -> 
                if sym#in_closure then
                  (c, Ksi.Comment(sym#get_name ^ " in closure", Ksi.Nop))
                else
                  (c, Ksi.Var(ksi_local_type sym#get_type, 
                              ksi_name sym#get_name))
            | None -> assert false)
      | Pair (a, b) ->
          let (c', a') = ksi_stmt (c, a) in
          let (c'', b') = ksi_stmt (c', b) in
          (c'', Ksi.Compound(a', b'))
      | If(e, s1, s2) ->
          let (c1, e') = ksi_exp (c, e) in
          let (c2, s1') = ksi_stmt (c1, s1) in
          let (c3, s2') = ksi_stmt (c2, s2) in
          (* Convince Ksi typechcecker we know, what we're doing... 
           * Otherwise it will complaing if types of if branch doesn't
           * agree... *)
          let s1'' = Ksi.Compound(s1', Ksi.Nop) in
          let s2'' = 
            match s2' with
                Ksi.Nop -> s2'
              | _ -> Ksi.Compound(s2', Ksi.Nop) in
          (c3, Ksi.Cond(e', s1'', s2''))
      | BreakIfNot(e) ->
          let (c', e') = ksi_exp (c, e) in
          (c', Ksi.ExitLoop(Ksi.Unary(Ksi.TruthNot, e')))
      | Loop(s) ->
          let (c', s') = ksi_stmt (c, s) in
          (c', Ksi.Loop(s'))
      | Open _ -> (c, Ksi.Nop) in
   (c', Ksi.Line(s.stmt_loc.loc_line, s'))
        
and ksi_exp (c, e) =
  let rec ksi_call_parms (c, p) =
    (* first, try to create temps for floats *)
    let rec mkfloats no r c p =
      match p with
          x::xs ->
            (match type_of x with
                Float -> 
                  let (c', e') = ksi_exp (c, x) in
                  let tmp_name = ksi_name ("_" ^ string_of_int(no) ^ "$") in
                  let v = Ksi.Var(ksi_local_type Float, tmp_name) in
                  let vi = Ksi.Modify(Ksi.Ref(tmp_name), e') in
                  mkfloats (no+1) (vi::v::r) c' xs
              | _ -> mkfloats no r c xs)
        | [] -> (c, List.rev r) in
    (* then compute other parms, (using temps for floats) *)
    let rec mkexps no r c p =
      match p with
          x::xs ->
            (match type_of x with
                Float -> 
                  let tmp_name = ksi_name ("_" ^ string_of_int(no) ^ "$") in
                  let v = Ksi.Ref(tmp_name) in
                  mkexps (no+1) (v::r) c xs
              | _ ->
                let (c', e') = ksi_exp (c, x) in
                mkexps no (e'::r) c' xs)
        | [] -> (c, List.rev r) in
    let (c', tmps) = mkfloats 0 [] c p in
    let (c'', parms) = mkexps 0 [] c' p in
    (c'', tmps, parms) in
  let toplev_func sym =
    match instance_value sym#get_type with
        Func _ when sym#get_scope_level <= 0 -> true
      | _ -> false in
  let (c', e') =
    match e.raw_exp with
        SimpleBinOp(Assign, e1, e2) ->
          let (c1, e1') = ksi_exp (c, e1) in
          let (c2, e2') = ksi_exp (c1, e2) in
          (c2, Ksi.Modify(e1', e2'))
      | SimpleBinOp(op, e1, e2) -> 
          let (c1, e1') = ksi_exp (c, e1) in
          let (c2, e2') = ksi_exp (c1, e2) in
          (c2, Ksi.Binary(ksi_op2 op, e1', e2'))
      | SimpleUnOp(op, e) -> 
          let (c', e') = ksi_exp (c, e) in
          (c', Ksi.Unary(ksi_op1 op, e'))
      | Ref {ref_symbol = Some sym}
      | CrossModuleRef {xmr_symbol = Some sym}
                when (toplev_func sym) ->
        (* Need to encapsulate toplevel function into a function with 
         * closure. *)
        let aux_name = sym#get_name ^ "_" ^ (seq_name "closurized") ^ "$" in
        let (ret_type, parm_types) = 
          match instance_value sym#get_type with
              Func {ft_ret_type=r; ft_parms=p} -> (r, p)
            | _ -> assert false in
        let cnt = ref 0 in
        let parm_names = 
          (List.map (fun _ -> incr cnt; 
                              ("_" ^ string_of_int(!cnt))) parm_types) in
        let aux = 
           Ksi.Func([Ksi.Local], ksi_passed_type ret_type, ksi_name aux_name, 
                    (Ksi.Ptr(Ksi.Void), ksi_name "dummy_cl$") :: 
                      (List.map2 (fun t n -> (ksi_passed_type t, ksi_name n)) 
                                 parm_types parm_names),
                    Ksi.Return(Ksi.Call(ksi_name sym#get_name, 
                      (List.map (fun n -> Ksi.Ref(ksi_name n)) 
                                parm_names)))) in
        let lambda_val = 
           Ksi.Call(ksi_name "make_lambda$", 
                    [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                                 Ksi.FuncRef(ksi_name aux_name));
                     Ksi.Null(Ksi.Ptr(Ksi.Void))]) in
        ({c with ctx_aux_decls = aux :: c.ctx_aux_decls}, lambda_val)
      | CrossModuleRef {xmr_symbol = Some sym} ->
          (c, Ksi.Ref(ksi_name sym#get_name))
      | CrossModuleRef _ -> assert false
      | Ref ({ref_symbol = Some sym} as r) ->
          let code = 
            let rec refer lev =
              if lev == 1 then 
                 ksi_parent_closure
              else 
                 Ksi.Arrow(refer (lev-1), ksi_name "parent_cl$") in
            if r.ref_level == 0 then
              if sym#in_closure then
                ksi_closure_var sym#get_name
              else
                Ksi.Ref(ksi_name sym#get_name)
            else
              (assert sym#in_closure;
               Ksi.Arrow(refer r.ref_level, ksi_name sym#get_name)) in
          let code' = 
             match instance_value sym#get_type with
                 Bool -> Ksi.Binary(Ksi.Ne, code, 
                                    Ksi.IntConst(Ksi.ptroff (), "0"))
               | _ -> code in
          (c, code')
      | Ref _ -> assert false
      (* The easy case -- global function call. *)
      | Call({raw_exp = Ref {ref_symbol = Some sym}}, p) 
        when (toplev_func sym) ->
          let (c', t, p') = ksi_call_parms (c, p) in
          let call = Ksi.Call(ksi_name sym#get_name, p') in
          let call' = 
             match instance_value sym#get_type with
                 Func {ft_ret_type=Float} -> Ksi.Indirect(call)
               | _ -> call in
          let code = 
            match t with
                [] -> call'
              | _ -> Ksi.Bind(Ksi.CompoundList(t @ [call'])) in
          (c', code)
      (* The hard case -- function call with closure. *)
      | Call(e, p) ->
          let (c', e') = ksi_exp (c, e) in
          let (c'', t, p') = ksi_call_parms (c', p) in
          let (ret_type, parm_types) =
            match type_of e with
                Func {ft_ret_type=r; ft_parms=p} -> (r, p)
              | _ -> assert false in
          let lambda_var = 
             Ksi.Var(Ksi.Ptr(Ksi.Struct (ksi_name "lambda_exp$")), 
                     ksi_name "lambda$") in
          let lambda_init =
             Ksi.Modify(Ksi.Ref(ksi_name "lambda$"), e') in
          let lambda_fld n = 
             Ksi.Arrow(Ksi.Ref(ksi_name "lambda$"), ksi_name n) in
          let ksi_fnc_type = 
             Ksi.FuncPtr(ksi_passed_type ret_type,
                         Ksi.Ptr(Ksi.Void) :: 
                         (List.map ksi_passed_type parm_types)) in
          let fnc_ptr = Ksi.Convert(ksi_fnc_type, lambda_fld "fnc") in
          let call = Ksi.PtrCall(fnc_ptr, (lambda_fld "closure") :: p') in
          let call' =
             match ret_type with
                 Float -> Ksi.Indirect(call)
               | _ -> call in
          (c'', Ksi.Bind(Ksi.CompoundList(lambda_var :: 
                                          lambda_init :: t @ [call'])))
      | IntConst(i) -> (c, Ksi.IntConst(ksi_local_type Int, i))
      | BoolConst(b) -> (c, Ksi.BoolConst(b))
      | FloatConst(f) -> (c, Ksi.FloatConst(ksi_local_type Float, f))
      | StringConst(s) -> 
         (c, Ksi.Call(ksi_name "make_string$", 
                      [Ksi.IntConst(Ksi.uptroff(), 
                                    string_of_int(String.length s));
                       Ksi.StringConst(s)]))
      | Lambda {lambda_parms=p; lambda_ret_type=t; lambda_body=s} ->
        let aux_name = (seq_name "lambda") ^ "$" in
        let aux_decls = ksi_fnc_decl [Ksi.Local] c (t, p) aux_name s in
        let lambda_val = 
          Ksi.Call(ksi_name "make_lambda$", 
                   [Ksi.Convert(Ksi.Ptr(Ksi.Void), 
                                Ksi.FuncRef(ksi_name aux_name));
                    if c.ctx_closure_type = "" then
                      Ksi.Null(Ksi.Ptr(Ksi.Void))
                    else
                      ksi_closure]) in
        ({c with ctx_aux_decls = aux_decls @ c.ctx_aux_decls}, lambda_val) in
  (c', Ksi.Line(e.exp_loc.loc_line, e'))
  
and ksi_fnc_decl storage c1 (ret_type, parms) fnname body =
  let closure_struct name cl =
    let flds = (List.map (fun sym -> (ksi_passed_type sym#get_type, 
                                      ksi_name sym#get_name)) cl) in
    match c1.ctx_closure_type with
        "" -> Ksi.StructDef(name, [], flds)
      | _ -> let parent_cl = (Ksi.Ptr(Ksi.Struct(ksi_name c1.ctx_closure_type)), 
                              ksi_name "parent_cl$") in
             Ksi.StructDef(name, [], parent_cl :: flds) in
  let (need_cl, cl) = closure (c1.ctx_level + 1) body in
  let c2 = {ctx_level = c1.ctx_level + 1;
            ctx_closure_type = "";
            ctx_aux_decls = [];
            ctx_fnname = fnname;
            ctx_module = c1.ctx_module} in
  let c3 = if not need_cl then c2 else
           let name = fnname ^ "_" ^ (seq_name "closure") ^ "_t$" in
               {c2 with
                ctx_closure_type = name;
                ctx_aux_decls = (closure_struct (ksi_name name) cl) :: 
                                c2.ctx_aux_decls} in
  let (c4, code) = ksi_stmt (c3, body) in
  let code_with_cl =
    if not need_cl then 
      code 
    else
      let cl_struct = Ksi.Struct(ksi_name c3.ctx_closure_type) in
      let cl_var = Ksi.Var(Ksi.Ptr(cl_struct), ksi_name "cl$") in
      let cl_init = 
          Ksi.Modify(ksi_closure,
                     Ksi.Convert(Ksi.Ptr(cl_struct),
                                 Ksi.Call(ksi_name "make_closure$",
                                          [Ksi.SizeOf(cl_struct)]))) in
      let cl_copy_parent =
        match c1.ctx_closure_type with
            "" -> Ksi.Nop
          | _ -> Ksi.Modify(ksi_closure_var "parent_cl$",
                            ksi_parent_closure) in
      let cl_copy_parms =
         let rec helper parms =
           match parms with
               {parm_symbol = Some sym}::xs ->
                 let rest = helper xs in
                 if sym#in_closure then
                   Ksi.Modify(ksi_closure_var sym#get_name, 
                              Ksi.Ref(ksi_name sym#get_name)) :: rest
                 else
                    rest
             | [] -> []
             | _ -> assert false in
          helper parms in
      Ksi.CompoundList([cl_var; cl_init; cl_copy_parent] @ 
                       cl_copy_parms @ 
                       [code]) in
  let (ksi_ret_type, ksi_parms) = 
         (ksi_passed_type ret_type, 
          (List.map (fun {parm_type=t; parm_name=n} -> 
                                  (ksi_passed_type t, ksi_name n))
                    parms)) in
  let ksi_parms' =
    if c1.ctx_closure_type <> "" then
      (Ksi.Ptr(Ksi.Struct(ksi_name c1.ctx_closure_type)), 
       ksi_name "parent_cl$") 
      :: ksi_parms
    else if c1.ctx_level > 0 then
      (Ksi.Ptr(Ksi.Void), ksi_name "dummy_cl$") :: ksi_parms
    else
      ksi_parms in
  let func_decl = 
    Ksi.Func(storage,
             ksi_ret_type,
             ksi_name fnname,
             ksi_parms',
             code_with_cl) in
  match c4.ctx_aux_decls with
      [] -> 
        [Ksi.TopComment("END " ^ fnname); 
         func_decl; 
         Ksi.TopComment("BEG " ^ fnname)]
    | aux -> 
        Ksi.TopComment("END " ^ fnname) ::
        func_decl ::
        Ksi.TopComment("BEG " ^ fnname) :: 
        aux @ 
        [Ksi.TopComment("AUX " ^ fnname)]
;;

let wrap_as name sym =
  let (ret_type, parm_types) = 
    match instance_value sym#get_type with
        Func {ft_ret_type=r; ft_parms=p} -> (r, p)
      | _ -> assert false in
  let cnt = ref 0 in
  let parm_names = 
    (List.map (fun _ -> incr cnt; 
                        ("_" ^ string_of_int(!cnt))) parm_types) in
   Ksi.Func([Ksi.Public], ksi_passed_type ret_type, ksi_name name, 
            (List.map2 (fun t n -> (ksi_passed_type t, ksi_name n)) 
                       parm_types parm_names),
            Ksi.Return(Ksi.Call(ksi_name sym#get_name, 
              (List.map (fun n -> Ksi.Ref(ksi_name n)) parm_names)))) 
;;

let ksi_code ctx d =
  let (ctx', decls) = 
    match d with
        FuncDecl ({fn_symbol = Some sym; fn_loc = l} as f) ->
          let c = {global_ctx with ctx_module = ctx.gctx_module} in
          let (ctx', storage) = global_check l ctx sym in
          let decls = (ksi_fnc_decl storage c
                                    (f.fn_ret_type, f.fn_parms)
                                    sym#get_name 
                                    f.fn_body) in
          if f.fn_export_as = "" then
            (ctx', List.rev decls)
          else
            (ctx', List.rev (wrap_as f.fn_export_as sym :: decls))
      | FuncDecl _ -> assert false
      | GlobalDecl {global_symbol=Some sym; global_loc = l} ->
           let (ctx', storage) = (global_check l ctx sym) in
           (ctx', [Ksi.Global(storage,
                              ksi_local_type sym#get_type,
                              ksi_name sym#get_name)])
      | GlobalDecl _ -> assert false
      | ExternC {xc_symbol = Some sym} ->
         (ctx,
           (match instance_value sym#get_type with
               Func {ft_ret_type=r; ft_parms=a} ->
                 [Ksi.FuncProto([Ksi.Extern], ksi_passed_type r,
                                ksi_name sym#get_name,
                                (List.map ksi_passed_type a))]
             | t ->
                 [Ksi.Global([Ksi.Extern], ksi_passed_type t, 
                             ksi_name sym#get_name)]))
      | ExternC _ -> assert false
      | FuncProto _ -> assert false
      | TopLevelOpen _ -> (ctx, []) 
      | TypeDef _ -> (ctx, []) (* FIXME *) in
   let decls' = (List.rev !aux_decls) @ decls in
   aux_decls := [];
   (ctx', decls')
;;

let check_module ctx =
  match ctx.gctx_globals with
      [] -> ()
    | l -> 
      let oops = String.concat ", " (List.map (fun (t, n) -> n) l) in
      raise (Error ("following symbols, required by interface, are not \
                     provided: " ^ oops))
;;
